from copy import copy
from dataclasses import dataclass, fields
from typing import Optional, Any

from dataclasses_json import DataClassJsonMixin, config, Undefined

# Ensures we don't render 'null' fields.
ignore_none_cfg = config(undefined=Undefined.EXCLUDE, exclude=lambda f: f is None)["dataclasses_json"]


@dataclass
class Op:
    DELETE: str = "del"
    INIT: str = "init"
    UPSERT: str = "ups"
    ARR_INSERT: str = "arr_ins"


_op_fields = [f.default for f in fields(Op)]


@dataclass
class Change(DataClassJsonMixin):
    dataclass_json_config = copy(ignore_none_cfg)
    op: str
    path: str  # jsonpath str
    value: Optional[Any] = None  # Optional when deleting
    index: Optional[int] = None  # Used only in list processing

    def __post_init__(self):
        if self.op not in _op_fields:
            raise ValueError("Operation '%s' not found in list of legal operations, %s" % (self.op, _op_fields))


@dataclass
class ChangeSet(DataClassJsonMixin):
    changes: list[Change]



import json
from copy import copy
from typing import Union

from .models import ChangeSet, Op
from jsonpath_ng import Fields, Index
from jsonpath_ng.ext import parse


def update_json(input: Union[str, dict], changeset: Union[str, ChangeSet]) -> Union[str, dict]:
    """
    Updates a json (or json-compatible dict) as per the provided changeset, returning the modified result.

    :param input: If passed a json str, will parse into dict
    :param changeset: If passed a json str, will parse into ChangeSet object
    :return: will return whatever format was passed in as input
    """

    input_type_json = False
    if type(input) == str:
        input = json.loads(input)
        input_type_json = True
    elif type(input) == dict:
        input = copy(input)
    else:
        raise ValueError("Unsupported input type %s" % type(input))

    if type(changeset) == str:
        changeset = ChangeSet.from_json(changeset)
    elif type(changeset) == dict:
        changeset = ChangeSet.from_dict(changeset)
    elif type(changeset) != ChangeSet:
        raise ValueError("Unsupported ChangeSet type %s" % type(input))

    for change in changeset.changes:
        expr = parse(change.path)

        if change.op == Op.DELETE:
            idx_item = []
            for m in expr.find(input):
                if type(m.path) == Fields:
                    fields = list(filter(lambda p: p is not None, m.path.reified_fields(input)))
                    for field in fields:
                        del m.context.value[field]
                elif type(m.path) == Index:  # If we have an array element to delete
                    idx_item.append((m.path.index, m.context))

        elif change.op == Op.INIT:
            for m in expr.find_or_create(input):
                if type(m.path) == Fields:
                    fields = list(filter(lambda p: p is not None, m.path.reified_fields(input)))
                    for field in fields:
                        if m.context.value[field] == {}:
                            m.context.value[field] = change.value
                elif type(m.path) == Index:  # If we have an array element to delete
                    if m.context.value[m.path.index] == {}:
                        m.context.value[m.path.index] = change.value

            for index, context in sorted(idx_item, key=lambda ix: ix[0], reverse=True):
                if context.value:
                    context.value.pop(index)  # always pop the last item in each

        elif change.op == Op.REPLACE:
            for m in expr.find(input):
                fields = list(filter(lambda p: p is not None, m.path.reified_fields(input)))
                for field in fields:
                    m.context.value[field] = change.value

        elif change.op == Op.UPSERT:
            input = expr.update_or_create(input, change.value)

        elif change.op == Op.ARR_INSERT:
            # Only for array insertion
            for m in expr.find(input):
                if type(m.value) != list:
                    raise ValueError("Operator '%s' should only be used against arrays, but was aimed at a '%s'" % (Op.ARR_INSERT, type(m.value)))
                else:
                    if change.index is None:
                        m.value.append(change.value)
                    elif type(change.index) == int:
                        m.value.insert(change.index, change.value)
                    else:
                        raise ValueError('Array Index of "%s" was passed, which was not of type int' % change.index)


    # Convert back to json str if necessary.
    if input_type_json:
        input = json.dumps(input)

    return input



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="json-updater",
    version="0.0.5",
    author="Craig McConomy",
    author_email="cmcconomy@fwig.com",
    description="Updates JSON according to a change spec",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FortressEngineering/json-updater",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 1 - Planning",
        "Operating System :: POSIX :: Linux"
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.9",
    install_requires=['jsonpath-ng ~= 1.5.3', 'dataclasses-json ~= 0.5.7']
)